import json

import bioc


def parse_collection(obj):
    collection = bioc.BioCCollection()
    collection.source = obj['source']
    collection.date = obj['date']
    collection.key = obj['key']
    collection.infons = obj['infons']
    for doc in obj['documents']:
        collection.add_document(parse_doc(doc))
    return collection


def parse_annotation(obj):
    ann = bioc.BioCAnnotation()
    ann.id = obj['id']
    ann.infons = obj['infons']
    ann.text = obj['text']
    for loc in obj['locations']:
        ann.add_location(bioc.BioCLocation(loc['offset'], loc['length']))
    return ann


def parse_relation(obj):
    rel = bioc.BioCRelation()
    rel.id = obj['id']
    rel.infons = obj['infons']
    for node in obj['nodes']:
        rel.add_node(bioc.BioCNode(node['refid'], node['role']))
    return rel


def parse_sentence(obj):
    sentence = bioc.BioCSentence()
    sentence.offset = obj['offset']
    sentence.infons = obj['infons']
    sentence.text = obj['text']
    for annotation in obj['annotations']:
        sentence.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        sentence.add_relation(parse_relation(relation))
    return sentence


def parse_passage(obj):
    passage = bioc.BioCPassage()
    passage.offset = obj['offset']
    passage.infons = obj['infons']
    if 'text' in obj:
        passage.text = obj['text']
    for sentence in obj['sentences']:
        passage.add_sentence(parse_sentence(sentence))
    for annotation in obj['annotations']:
        passage.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        passage.add_relation(parse_relation(relation))
    return passage


def parse_doc(obj):
    doc = bioc.BioCDocument()
    doc.id = obj['id']
    doc.infons = obj['infons']
    for passage in obj['passages']:
        doc.add_passage(parse_passage(passage))
    for annotation in obj['annotations']:
        doc.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        doc.add_relation(parse_relation(relation))
    return doc


def load(fp, **kwargs):
    """
    Deserialize fp (a .read()-supporting text file or binary file containing a JSON document) to a BioCCollection object

    Args:
        fp: a file containing a JSON document
        **kwargs:

    Returns:
        BioCCollection: a collection
    """
    obj = json.load(fp, **kwargs)
    return parse_collection(obj)


def loads(s, **kwargs):
    """
    Deserialize s (a str, bytes or bytearray instance containing a JSON document) to a BioCCollection object.

    Args:
        s(str):
        **kwargs:

    Returns:
        BioCCollection: a collection
    """
    obj = json.loads(s, **kwargs)
    return parse_collection(obj)



import lxml.etree as etree
from .bioc import BioCCollection
from .encoder import encode_infon, encode_document


class BioCEncoderIter(object):
    def __writer__(self):
        with etree.xmlfile(self.file, encoding=self.collection.encoding, close=True) as xf:
            xf.write_declaration(standalone=self.collection.standalone)
            with xf.element('collection'):
                try:
                    while True:
                        el = (yield)
                        xf.write(el)
                        xf.write('\n')
                        xf.flush()
                except GeneratorExit:
                    pass

    def __init__(self, name, collection=None):
        """
        Returns an object of the BioCEncoderIter which can write an BioC file
        incrementally at document level.

        Args:
            name(str): file name to be decode
        """

        self.file = name
        if not collection:
            collection = BioCCollection()
        self.collection = collection
        self.w = self.__writer__()
        next(self.w)   # start writing (run up to 'yield')

        elem = etree.Element('source')
        elem.text = self.collection.source
        self.w.send(elem)

        elem = etree.Element('date')
        elem.text = self.collection.date
        self.w.send(elem)

        elem = etree.Element('key')
        elem.text = self.collection.key
        self.w.send(elem)

        for k, v in self.collection.infons.items():
            elem = encode_infon(k, v)
            self.w.send(elem)

    def close(self):
        self.w.close()

    def writedocument(self, document):
        tree = encode_document(document)
        self.w.send(tree)




import lxml.etree as etree
import io

from .bioc import (
    BioCCollection,
    BioCDocument,
    BioCPassage,
    BioCSentence,
    BioCAnnotation,
    BioCRelation,
    BioCLocation,
    BioCNode
)


class BioCDecoder(object):
    def __init__(self):
        pass

    def decodes(self, s):
        """
        Deserialize ``s`` to a BioC collection object.

        Args:
            s(str): a "str" instance containing a BioC collection

        Returns:
            BioCCollection: a object of BioCollection
        """
        tree = etree.parse(io.BytesIO(bytes(s, encoding='UTF-8')))
        collection = self.__parse_collection(tree.getroot())
        collection.encoding = tree.docinfo.encoding
        collection.standalone = tree.docinfo.standalone
        collection.version = tree.docinfo.xml_version
        return collection

    def decode(self, fp):
        """
        Deserialize ``fp`` to a BioC collection object.

        Args:
            fp: a ``.read()``-supporting file-like object containing a BioC collection

        Returns:
            BioCCollection: a object of BioCollection
        """
        # utf8_parser = etree.XMLParser(encoding='utf-8')
        tree = etree.parse(fp)
        collection = self.__parse_collection(tree.getroot())
        collection.encoding = tree.docinfo.encoding
        collection.standalone = tree.docinfo.standalone
        collection.version = tree.docinfo.xml_version
        return collection

    def __parse_collection(self, tree):
        collection = BioCCollection()
        collection.source = tree.findtext('source')
        collection.date = tree.findtext('date')
        collection.key = tree.findtext('key')
        collection.infons = self.__parse_infons(tree)
        for child in tree.findall('document'):
            collection.add_document(self.__parse_document(child))
        return collection

    def __parse_document(self, tree):
        document = BioCDocument()
        document.id = tree.findtext('id')
        document.infons = self.__parse_infons(tree)
        for child in tree.findall('passage'):
            document.add_passage(self.__parse_passage(child))
        for child in tree.findall('annotation'):
            document.add_annotation(self.__parse_annotation(child))
        for child in tree.findall('relation'):
            document.add_relation(self.__parse_relation(child))
        return document

    def __parse_passage(self, tree):
        passage = BioCPassage()
        passage.offset = int(tree.findtext('offset'))
        passage.infons = self.__parse_infons(tree)
        if tree.find('text') is not None:
            passage.text = tree.findtext('text')
        for child in tree.findall('sentence'):
            passage.add_sentence(self.__parse_sentence(child))
        for child in tree.findall('annotation'):
            passage.add_annotation(self.__parse_annotation(child))
        for child in tree.findall('relation'):
            passage.add_relation(self.__parse_relation(child))
        return passage

    def __parse_sentence(self, tree):
        sentence = BioCSentence()
        sentence.offset = int(tree.findtext('offset'))
        sentence.text = tree.findtext('text')
        sentence.infons = self.__parse_infons(tree)
        for child in tree.findall('annotation'):
            sentence.add_annotation(self.__parse_annotation(child))
        for child in tree.findall('relation'):
            sentence.add_relation(self.__parse_relation(child))
        return sentence

    def __parse_annotation(self, tree):
        annotation = BioCAnnotation()
        annotation.id = tree.attrib['id']
        annotation.infons = self.__parse_infons(tree)
        annotation.text = tree.findtext('text')
        for child in tree.findall('location'):
            annotation.add_location(
                BioCLocation(int(child.attrib['offset']), int(child.attrib['length'])))
        return annotation

    def __parse_relation(self, tree):
        relation = BioCRelation()
        if 'id' in tree.attrib:
            relation.id = tree.attrib['id']
        relation.infons = self.__parse_infons(tree)
        for child in tree.findall('node'):
            relation.add_node(BioCNode(child.attrib['refid'], child.attrib['role']))
        return relation

    @classmethod
    def __parse_infons(cls, tree):
        return {child.attrib['key']: child.text for child in tree.findall('infon')}



import json

import bioc


def dumps(obj, **kwargs):
    """
    Serialize a BioC ``obj`` to a JSON formatted ``str``.
    """
    return json.dumps(obj, cls=BioCJSONEncoder, **kwargs)


def dump(obj, fp, **kwargs):
    """
    Serialize obj as a JSON formatted stream to fp (a .write()-supporting file-like object)
    """
    return json.dump(obj, fp, cls=BioCJSONEncoder, **kwargs)


class BioCJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bioc.BioCLocation):
            return {
                'offset': obj.offset,
                'length': obj.length,
            }
        elif isinstance(obj, bioc.BioCAnnotation):
            return {
                'id': obj.id,
                'infons': obj.infons,
                'text': obj.text,
                'locations': [self.default(l) for l in obj.locations],
            }
        elif isinstance(obj, bioc.BioCNode):
            return {
                'refid': obj.refid,
                'role': obj.role,
            }
        elif isinstance(obj, bioc.BioCRelation):
            return {
                'id': obj.id,
                'infons': obj.infons,
                'nodes': [self.default(n) for n in obj.nodes]
            }
        elif isinstance(obj, bioc.BioCSentence):
            return {
                'offset': obj.offset,
                'infons': obj.infons,
                'text': obj.text,
                'annotations': [self.default(a) for a in obj.annotations],
                'relations': [self.default(r) for r in obj.relations],
            }
        elif isinstance(obj, bioc.BioCPassage):
            return {
                'offset': obj.offset,
                'infons': obj.infons,
                'text': obj.text,
                'sentences': [self.default(s) for s in obj.sentences],
                'annotations': [self.default(a) for a in obj.annotations],
                'relations': [self.default(r) for r in obj.relations],
            }
        elif isinstance(obj, bioc.BioCDocument):
            return {
                'id': obj.id,
                'infons': obj.infons,
                'passages': [self.default(p) for p in obj.passages],
                'annotations': [self.default(a) for a in obj.annotations],
                'relations': [self.default(r) for r in obj.relations],
            }
        elif isinstance(obj, bioc.BioCCollection):
            return {
                'source': obj.source,
                'date': obj.date,
                'key': obj.key,
                'infons': obj.infons,
                'documents': [self.default(d) for d in obj.documents],
            }
        return json.JSONEncoder.default(self, obj)



import bioc

DOCUMENT = 1
PASSAGE = 2
SENTENCE = 3


def annotations(obj, docid=None, level=PASSAGE):
    """
    Get all annotations in document id.

    Args:
        obj: BioCCollection, BioCDocument, BioCPassage, or BioCSentence
        docid(str): document id. If None, all documents
        level(int): one of DOCUMENT, PASSAGE, SENTENCE

    Yields:
        BioCAnnotation: one annotation
    """
    if isinstance(obj, bioc.BioCCollection):
        for document in filter(lambda d: docid is None or docid == d.id, obj.documents):
            for ann in annotations(document, level=level):
                yield ann
    elif isinstance(obj, bioc.BioCDocument):
        if level == DOCUMENT:
            for ann in obj.annotations:
                yield ann
        elif level == PASSAGE or level == SENTENCE:
            for passage in obj.passages:
                for ann in annotations(passage, level=level):
                    yield ann
        else:
            raise ValueError('level must be DOCUMENT, PASSAGE, or SENTENCE')
    elif isinstance(obj, bioc.BioCPassage):
        if level == PASSAGE:
            for ann in obj.annotations:
                yield ann
        elif level == SENTENCE:
            for sentence in obj.sentences:
                for ann in annotations(sentence, level=level):
                    yield ann
        else:
            raise ValueError('level must be SENTENCE')
    elif isinstance(obj, bioc.BioCSentence):
        if level == SENTENCE:
            for ann in obj.annotations:
                yield ann
        else:
            raise ValueError('level must be SENTENCE')
    else:
        raise ValueError('obj must be BioCCollection, BioCDocument, BioCPassage, or BioCSentence')


def relations(obj, docid=None, level=PASSAGE):
    """
    Get all relations in document id.

    Args:
        obj: BioCCollection, BioCDocument, BioCPassage, or BioCSentence
        docid(str): document id. If None, all documents
        level(int): one of DOCUMENT, PASSAGE, SENTENCE

    Yields:
        BioCRelation: one relation
    """
    if isinstance(obj, bioc.BioCCollection):
        for document in filter(lambda d: docid is None or docid == d.id, obj.documents):
            for rel in relations(document, level=level):
                yield rel
    elif isinstance(obj, bioc.BioCDocument):
        if level == DOCUMENT:
            for rel in obj.relations:
                yield rel
        elif level == PASSAGE or level == SENTENCE:
            for passage in obj.passages:
                for rel in relations(passage, level=level):
                    yield rel
        else:
            raise ValueError('level must be DOCUMENT, PASSAGE, or SENTENCE')
    elif isinstance(obj, bioc.BioCPassage):
        if level == PASSAGE:
            for rel in obj.relations:
                yield rel
        elif level == SENTENCE:
            for sentence in obj.sentences:
                for rel in relations(sentence, level=level):
                    yield rel
        else:
            raise ValueError('level must be SENTENCE')
    elif isinstance(obj, bioc.BioCSentence):
        if level == SENTENCE:
            for rel in obj.relations:
                yield rel
        else:
            raise ValueError('level must be SENTENCE')
    else:
        raise ValueError('obj must be BioCCollection, BioCDocument, BioCPassage, or BioCSentence')


def sentences(obj):
    """
    Get all sentences in document id.

    Args:
        obj: BioCCollection, BioCDocument, or BioCPassage

    Yields:
        BioCSentence: one sentence
    """
    if isinstance(obj, bioc.BioCCollection):
        for document in obj.documents:
            for sentence in sentences(document):
                yield sentence
    elif isinstance(obj, bioc.BioCDocument):
        for passage in obj.passages:
            for sentence in sentences(passage):
                yield sentence
    elif isinstance(obj, bioc.BioCPassage):
        for sentence in obj.sentences:
            yield sentence
    else:
        raise ValueError('obj must be BioCCollection, BioCDocument, or BioCPassage')



import lxml.etree as etree

from .bioc import (
    BioCCollection,
    BioCDocument,
    BioCPassage,
    BioCSentence,
    BioCAnnotation,
    BioCRelation,
    BioCLocation,
    BioCNode
)


class BioCDecoderIter(object):
    """
    Code to read/write BioC XML in an incremental way.
    """

    def __init__(self, name):
        """
        Open an object of the BioCDecoderIter which can parse an BioC file
        incrementally at document level.

        Args:
            name(str): file name to be decoded
        """
        self.file = name
        self.__context = iter(etree.iterparse(self.file, events=('start', 'end')))
        self.__state = 0
        self.__read()

    def __iter__(self):
        return self

    def __next__(self):
        if self.__document is None:
            raise StopIteration
        else:
            document = self.__document
            self.__read()
            return document

    def next(self):
        """
        Reads one BioC document from the XML file.

        Returns:
            BioCDocument: the BioC document
        """
        return self.__next__()

    def __read(self):
        while self.__has_next():
            event, elem = self.__next_event()
            if self.__state == 0:
                if event == 'start':
                    if elem.tag == 'collection':
                        self.__state = 1
                        self.__collection = BioCCollection()
                        # collection information
            elif self.__state == 1:
                if event == 'start':
                    if elem.tag == 'document':
                        self.__document = BioCDocument()
                        self.__state = 2
                elif event == 'end':
                    if elem.tag == 'source':
                        self.__collection.source = elem.text
                    elif elem.tag == 'date':
                        self.__collection.date = elem.text
                    elif elem.tag == 'key':
                        self.__collection.key = elem.text
                    elif elem.tag == 'infon':
                        self.__collection.infons[elem.get('key')] = elem.text
                    elif elem.tag == 'collection':
                        self.__state = 0
                        self.__document = None
                        self.__passage = None
                        self.__sentence = None
            elif self.__state == 2:
                if event == 'start':
                    if elem.tag == 'passage':
                        self.__passage = BioCPassage()
                        self.__state = 3
                    elif elem.tag == 'annotation':
                        self.__document.add_annotation(self.__read_annotation(elem))
                    elif elem.tag == 'relation':
                        self.__document.add_relation(self.__read_relation(elem))
                elif event == 'end':
                    if elem.tag == 'id':
                        self.__document.id = elem.text
                    elif elem.tag == 'infon':
                        self.__document.infons[elem.get('key')] = elem.text
                    elif elem.tag == 'document':
                        self.__state = 1
                        return
            elif self.__state == 3:
                if event == 'start':
                    if elem.tag == 'sentence':
                        self.__sentence = BioCSentence()
                        self.__state = 4
                    elif elem.tag == 'annotation':
                        self.__passage.add_annotation(self.__read_annotation(elem))
                    elif elem.tag == 'relation':
                        self.__passage.add_relation(self.__read_relation(elem))
                elif event == 'end':
                    if elem.tag == 'offset':
                        self.__passage.offset = int(elem.text)
                    elif elem.tag == 'text':
                        self.__passage.text = elem.text
                    elif elem.tag == 'infon':
                        self.__passage.infons[elem.get('key')] = elem.text
                    elif elem.tag == 'passage':
                        self.__state = 2
                        if self.__passage is not None:
                            self.__document.add_passage(self.__passage)
            elif self.__state == 4:
                if event == 'start':
                    if elem.tag == 'annotation':
                        self.__sentence.add_annotation(self.__read_annotation(elem))
                    elif elem.tag == 'relation':
                        self.__sentence.add_relation(self.__read_relation(elem))
                elif event == 'end':
                    if elem.tag == 'offset':
                        self.__sentence.offset = int(elem.text)
                    elif elem.tag == 'text':
                        self.__sentence.text = elem.text
                    elif elem.tag == 'infon':
                        self.__sentence.infons[elem.get('key')] = elem.text
                    elif elem.tag == 'sentence':
                        self.__state = 3
                        if self.__sentence is not None:
                            self.__passage.add_sentence(self.__sentence)

    def __read_annotation(self, start_elem):
        ann = BioCAnnotation()
        ann.id = start_elem.get('id')
        while self.__has_next():
            event, elem = self.__next_event()
            if event == 'start':
                pass
            elif event == 'end':
                if elem.tag == 'text':
                    ann.text = elem.text
                elif elem.tag == 'infon':
                    ann.infons[elem.get('key')] = elem.text
                elif elem.tag == 'location':
                    ann.add_location(BioCLocation(int(elem.get('offset')), int(elem.get('length'))))
                elif elem.tag == 'annotation':
                    return ann
        raise RuntimeError("should not reach here")

    def __read_relation(self, start_elem):
        rel = BioCRelation()
        rel.id = start_elem.get('id')
        while self.__has_next():
            event, elem = self.__next_event()
            if event == 'start':
                pass
            elif event == 'end':
                if elem.tag == 'infon':
                    rel.infons[elem.get('key')] = elem.text
                elif elem.tag == 'node':
                    rel.add_node(BioCNode(elem.get('refid'), elem.get('role')))
                if elem.tag == 'relation':
                    return rel
        raise RuntimeError("should not reach here")

    def __has_next(self):
        try:
            self.__event, self.__elem = next(self.__context)
            return True
        except StopIteration:
            self.__event = None
            self.__elem = None
            return False

    def __next_event(self):
        return self.__event, self.__elem

    def get_collection_info(self):
        """
        Reads the collection information: encoding, version, DTD, source, date, key, infons, etc.

        Returns:
            BioCCollection: the BioC collection that contains only information
        """
        return self.__collection



from contextlib import contextmanager

from .bioc import (
    BioCCollection,
    BioCDocument,
    BioCPassage,
    BioCSentence,
    BioCAnnotation,
    BioCRelation,
    BioCLocation,
    BioCNode
)
from .decoder import BioCDecoder
from .encoder import BioCEncoder
from .iterdecoder import BioCDecoderIter
from .iterencoder import BioCEncoderIter
from .validator import BioCValidator
from .biocitertools import (
    PASSAGE,
    SENTENCE,
    DOCUMENT,
    annotations,
    relations,
    sentences,
)
from .utils import (
    get_text,
    pretty_print
)

from .jsonencoder import dumps as jsondumps
from .jsonencoder import dump as jsondump
from .jsondecoder import load as jsonload
from .jsondecoder import loads as jsonloads


__all__ = ['BioCAnnotation', 'BioCCollection', 'BioCDocument', 'BioCLocation', 'BioCNode',
           'BioCPassage', 'BioCRelation', 'BioCSentence', 'load', 'loads', 'dump', 'dumps',
           'iterparse', 'merge', 'validate', 'iterwrite', 'annotations', 'sentences', 'jsondumps', 'jsondump',
           'jsonload', 'jsonloads', 'get_text', 'pretty_print']


def dumps(collection, pretty_print=True):
    """
    Serialize ``collection`` to a BioC formatted ``str``.

    Args:
        collection (BioCollection): the BioC collection
        pretty_print (boolean): enables formatted XML

    Returns:
        str: a BioC formatted ``str``
    """
    return BioCEncoder(pretty_print=pretty_print).encode(collection)


def dump(collection, fp, pretty_print=True):
    """
    Serialize ``collection`` as a BioC formatted stream to ``fp``.

    Args:
        collection (BioCollection): the BioC collection
        fp: a ``.write()``-supporting file-like object
        pretty_print (boolean): enables formatted XML
    """
    fp.write(BioCEncoder(pretty_print=pretty_print).encode(collection))


def load(fp):
    """
    Deserialize ``fp`` to a BioC collection object.

    Args:
        fp: a ``.read()``-supporting file-like object containing a BioC collection

    Returns:
         BioCCollection: a object of BioCollection
    """
    return BioCDecoder().decode(fp)


def loads(s, encoding='UTF-8'):
    """
    Deserialize ``s`` to a BioC collection object.

    Args:
        s(str/bytes/bytearray): a ``str`` instance containing a BioC collection
        encoding(str): The input encoding should be UTF-8, UTF-16 or UTF-32.

    Returns:
        BioCCollection: a object of BioCollection
    """
    return BioCDecoder().decodes(s)


def validate(collection, onerror=None):
    BioCValidator(onerror).validate(collection)


def merge(dst, srcs):
    """
    Merge multiple BioC files into one.

    Args:
        dst(str): output BioC file name
        srcs(list): input BioC file names
    """
    collection = None
    for src in srcs:
        with open(src) as fp:
            tmp = load(fp)
        if collection is None:
            collection = tmp
        else:
            for document in tmp.documents:
                collection.add_document(document)
    with open(dst, 'w') as fp:
        dump(collection, fp)


@contextmanager
def iterparse(file):
    parser = BioCDecoderIter(file)
    yield parser


@contextmanager
def iterwrite(file, collection=None):
    writer = BioCEncoderIter(file, collection)
    yield writer
    writer.close()



import lxml.etree as etree
import bioc


def fill_char(text, offset, char='\n'):
    dis = offset - len(text)
    if dis < 0:
        raise ValueError
    if dis > 0:
        text += char * dis
    return text


def get_text(obj):
    """
    Return text with its offset in the document
    
    Args:
        obj: BioCDocument, BioCPassage, or BioCSentence
    
    Returns:
        tuple(int,str): offset, text
    """
    if isinstance(obj, bioc.BioCSentence):
        return obj.offset, obj.text,
    elif isinstance(obj, bioc.BioCPassage):
        if obj.text:
            return obj.offset, obj.text
        else:
            text = ''
            for sentence in obj.sentences:
                try:
                    text = fill_char(text, sentence.offset - obj.offset, ' ')
                    assert sentence.text, 'BioC sentence has no text: {}'.format(sentence.offset)
                    text += sentence.text
                except:
                    raise ValueError('Overlapping sentences %d' % (sentence.offset))
            return obj.offset, text
    elif isinstance(obj, bioc.BioCDocument):
        text = ''
        for passage in obj.passages:
            try:
                text = fill_char(text, passage.offset)
                text += get_text(passage)[1]
            except:
                raise ValueError('%s: overlapping passages %d' % (obj.id, passage.offset))
        return 0, text
    else:
        raise ValueError('obj must be BioCCollection, BioCDocument, BioCPassage, or BioCSentence')


def pretty_print(src, dst):
    """
    Pretty print the XML file
    """
    parser = etree.XMLParser(remove_blank_text=True)
    tree = etree.parse(src, parser)
    docinfo = tree.docinfo
    with open(dst, 'wb') as fp:
        fp.write(etree.tostring(tree, pretty_print=True,
                                encoding=docinfo.encoding, standalone=docinfo.standalone))




import lxml.etree as etree


class BioCEncoder(object):
    def __init__(self, pretty_print=True):
        """
        Constructor for BioCEncoder, with sensible defaults.

        Args:
            pretty_print(boolean): enables formatted XML
        """
        self.pretty_print = pretty_print

    def encode(self, collection):
        """
        Return a BioC formatted string representation of a BioC collection structure.

        Args:
            collection(BioCCollection): a BioC collection

        Returns:
            str: a BioC formatted ``str``
        """
        doc = etree.ElementTree(BioCEncoder.to_element_tree(collection))
        s = etree.tostring(doc, pretty_print=self.pretty_print, encoding=collection.encoding,
                           standalone=collection.standalone)
        return s.decode(collection.encoding)

    @classmethod
    def to_element_tree(cls, collection):
        """
        Returns the BioC xml tree that the BioCollection represents as a ElementTree.

        Args:
            collection(BioCCollection): a BioC collection

        Returns:
            ElementTree: the BioC xml that the BioCollection represents as a ElementTree.
        """
        tree = etree.Element('collection')
        etree.SubElement(tree, 'source').text = collection.source
        etree.SubElement(tree, 'date').text = collection.date
        etree.SubElement(tree, 'key').text = collection.key
        for k, v in collection.infons.items():
            tree.append(encode_infon(k, v))
        for d in collection.documents:
            tree.append(encode_document(d))
        return tree


def encode_document(document):
    tree = etree.Element('document')
    etree.SubElement(tree, 'id').text = document.id
    for k, v in document.infons.items():
        tree.append(encode_infon(k, v))
    for p in document.passages:
        tree.append(encode_passage(p))
    for a in document.annotations:
        tree.append(encode_annotation(a))
    for r in document.relations:
        tree.append(encode_relation(r))
    return tree


def encode_passage(passage):
    tree = etree.Element('passage')
    for k, v in passage.infons.items():
        tree.append(encode_infon(k, v))
    etree.SubElement(tree, 'offset').text = str(passage.offset)
    if passage.text:
        etree.SubElement(tree, 'text').text = passage.text
    for s in passage.sentences:
        tree.append(encode_sentence(s))
    for a in passage.annotations:
        tree.append(encode_annotation(a))
    for r in passage.relations:
        tree.append(encode_relation(r))
    return tree


def encode_sentence(sentence):
    tree = etree.Element('sentence')
    for k, v in sentence.infons.items():
        tree.append(encode_infon(k, v))
    etree.SubElement(tree, 'offset').text = str(sentence.offset)
    if sentence.text:
        etree.SubElement(tree, 'text').text = sentence.text
    for a in sentence.annotations:
        tree.append(encode_annotation(a))
    for r in sentence.relations:
        tree.append(encode_relation(r))
    return tree


def encode_annotation(annotation):
    tree = etree.Element('annotation', {'id': annotation.id})
    for k, v in annotation.infons.items():
        tree.append(encode_infon(k, v))
    for l in annotation.locations:
        etree.SubElement(tree, 'location', {'offset': str(l.offset), 'length': str(l.length)})
    etree.SubElement(tree, 'text').text = annotation.text
    return tree


def encode_relation(relation):
    tree = etree.Element('relation', {'id': relation.id})
    for k, v in relation.infons.items():
        tree.append(encode_infon(k, v))
    for n in relation.nodes:
        etree.SubElement(tree, 'node', {'refid': n.refid, 'role': n.role})
    return tree


def encode_infon(k, v):
    elem = etree.Element('infon', {'key': str(k)})
    elem.text = str(v)
    return elem
